-- REVIEW Document Displayer
-- by Matt Sephton
-- u5ms@csc.liv.ac.uk

--with profile
without type_check

include graphics.e
include image.e

integer iLines, iLinesR, iPos
sequence vc, sBuffer, sBookmarks, sSearch

vc = video_config()
constant iHeight = vc[VC_LINES] - 1
constant TRUE = 1, FALSE = 0
constant sVersion = "0.10"

function wait_key()
    return machine_func(26, 0)
end function

function min(integer first, integer second)
    if first <= second then
	return first
    else
	return second
    end if
end function

function max(integer first, integer second)
    if first >= second then
	return first
    else
	return second
    end if
end function

procedure show_status()
    position(vc[VC_LINES], 1)
    text_color(BRIGHT_WHITE)
    bk_color(BLUE)
    printf(1, " REVIEW lines %04d to %04d of %04d",
	      {iPos, min(iLinesR, iPos+(iHeight-1)), iLinesR})
    printf(1, "%46s"," ")
    text_color(WHITE)
    bk_color(BLACK)
end procedure

procedure message(sequence text)
    position(vc[VC_LINES], 1)
    text_color(BRIGHT_WHITE)
    bk_color(BLUE)
    puts(1, " REVIEW " & text & repeat(' ', vc[VC_COLUMNS] - length(text) - 8))
    position(vc[VC_LINES], length(text) + 10)
    text_color(WHITE)
    bk_color(BLACK)
end procedure

procedure readfile(sequence sFilename)
    integer fn
    object oLine
    
    -- read text file into a sequence
    message(sFilename)
    position(iHeight-1,1)
    puts(1, "By Matt Sephton (u5ms@csc.liv.ac.uk)")
    fn = open(sFilename, "r")
    sBuffer = {}
    while 1 do
	oLine = gets(fn)
	if atom(oLine) then
	    exit   -- end of file
	end if
	sBuffer = append(sBuffer, oLine)
    end while
    close(fn)
    iLinesR = length(sBuffer)
    iLines = iLinesR
    --docs less than a page
    if iLines < iHeight then
	for n = iLines+1 to iLines+iHeight-1 do
	    sBuffer = append(sBuffer, "\n")
	end for
    else
	--add blankpage to all docs
	for n = 1 to iHeight-1 do
	    sBuffer = append(sBuffer, "\n")
	end for
    end if
    iLines = length(sBuffer)
end procedure

procedure screen(integer start)
-- screen full of text from (start)
    if start != iPos then
	scroll(iHeight, 1, iHeight)
	position(1, 1)
	for i = start to start + (iHeight-1) do
	    puts(1, sBuffer[i])
	end for
	iPos = start
	show_status()
    end if
end procedure

procedure line(integer number, integer loc)
-- line (number) of text
    position(loc, 1)
    puts(1, sBuffer[number])
    show_status()
end procedure

procedure search(integer pos, sequence word)
-- search for (word) from (pos)
    message("Searching...")
    while match(word, sBuffer[pos]) = 0 and pos < iLines do
	pos = pos+1
    end while
    if match(word, sBuffer[pos]) > 0 then
	pos = max(1, min(pos, iLines-(iHeight-1)))
	screen(pos)
	text_color(BRIGHT_WHITE)
	line(iPos, 1)
	text_color(WHITE)
    end if
    show_status()
end procedure

function val(sequence string)
-- convert digit string to an integer
    atom n

    n = 0
    for i = 1 to length(string) do
	if string[i] >= '0' and string[i] <= '9' then
	    n = n * 10 + string[i] - '0'
	    if not integer(n) then
		return 0
	    end if
	else
	    exit
	end if
    end for
    return n
end function

function key_gets(sequence existing, sequence status, atom numeric)
-- return input string from keyboard
    sequence input_string
    integer line, init_column, column, char
    sequence cursorp

    message(status)
    text_color(BRIGHT_WHITE)
    bk_color(BLUE)
    cursor(#0607)
    cursorp = get_position()
    line = cursorp[1]
    init_column = cursorp[2]
    column = init_column
    
    -- initial input
    if length(existing) != 0 then
	input_string = existing
	puts(1, existing)
	column = column + length(existing)
    else
	input_string = ""
    end if
    
    while TRUE do
	char = wait_key()
	
	if char = 13 then
	    exit
	elsif char = 27 then
	    if column = init_column then
		exit
	    else
		column = init_column
		position(line, column)
		puts(1, repeat(' ', length(input_string)))
		position(line, column)
		input_string = ""
	    end if
	elsif char = 8 or char = 331 then
	    if column > init_column then
		column = column - 1
		position(line, column)
		puts(1, ' ')
		position(line, column)
		input_string = input_string[1..length(input_string)-1]
	    end if
	elsif char >= 32 or char = '\t' and char != 333 then
	    if column < vc[VC_COLUMNS] then
		if numeric and char >= '0' and char <= '9' or numeric = FALSE then
		    
		    if char = '\t' then
			puts(1, ' ')
		    else
			puts(1, char)
		    end if
		    column = column + 1
		    if column - init_column > length(input_string) then
			input_string = append(input_string, char)
		    else
			input_string[column - init_column] = char
		    end if
		end if
	    end if
	end if
    end while
    cursor(#2000)
    text_color(WHITE)
    bk_color(BLACK)
    if numeric then
	return val(input_string)
    else
	return input_string
    end if
end function

procedure init()
    sequence params
    
    wrap(0)
    cursor(#2000)
    params = command_line()

    --params = {"","",""}
    --params[3] = "Lines.txt"
    --params[3] = "Review.txt"
    --params[3] = "C:\\Windows\\Win.ini"
    if length(params) >= 3 then
	readfile(params[3])
    else
	puts(1, "Use: REVIEW [drive:\\path\\]filename.ext")
	abort(0)
    end if
    
    -- start at 0 lines
    iPos = 0
    sBookmarks = repeat(0, 9)
    sSearch = ""
end procedure

procedure help()
    sequence old_page
    
    old_page = save_text_image({1,1}, {vc[VC_LINES],vc[VC_COLUMNS]})
    scroll(iHeight, 1, iHeight)
    message("Help")
    position(1,1)
    text_color(BRIGHT_WHITE)
    bk_color(BLUE)
    puts(1, " REVIEW Document Displayer " & sVersion
	    & repeat(' ', 12) & "by Matt Sephton (u5ms@csc.liv.ac.uk) \n\n")
    text_color(WHITE)
    bk_color(BLACK)
    puts(1, "F1\t\tHelp\t\t\t\n\n")
    puts(1, "UP\t\tLine up\t\t\t\n")
    puts(1, "DOWN\t\tLine down\t\t\t\n")
    puts(1, "PAGE UP\t\tPage up\t\t\t\n")
    puts(1, "PAGE DOWN\tPage down\t\t\t\n")
    puts(1, "HOME\t\tTop\t\t\t\n")
    puts(1, "END\t\tBottom\t\t\t\n\n")
    puts(1, "F\t\tFind text\t\t\t\n")
    puts(1, "F\t\tFind next\t\t\t\n")
    puts(1, "G\t\tGoto line\t\t\t\n")
    puts(1, "ALT 1-9\t\tPlace bookmark\t\t\t\n")
    puts(1, "1-9\t\tGoto bookmark\t\t\t\n")
    puts(1, "\t\t\t\t\t\n")
    while get_key() = -1 do
    end while
    display_text_image({1,1}, old_page)
end procedure

procedure main()
    atom k

    init()

    screen(1)
    while TRUE do
	k = get_key()
	if k != -1 then
	    if k = 27 then                      -- esc
		exit
	    elsif k = 70 or k = 102 then        -- find
		sSearch = key_gets(sSearch, "Search for:", FALSE)
		if length(sSearch) != 0 then
		    search(1, sSearch)
		else
		    show_status()
		end if
	    elsif k = 78 or k = 110 then        -- next
		if length(sSearch) != 0 then
		    search(iPos+1, sSearch)
		end if
	    elsif k = 71 or k = 103 then        -- goto line
		k = key_gets("", "Goto line:", TRUE)
		if k > 0 and k <= iLines and iLines != iHeight then
		    screen(min(k, iLines-(iHeight-1)))
		else
		    show_status()
		end if
	    elsif k = 80 or k = 112 then        -- print
	    elsif k = 315 then                  -- F1 help
		help()
	    elsif k = 327 then                  -- home
		screen(1)
	    elsif k = 328 then                  -- up
		if iPos > 1 then
		    iPos = iPos - 1
		    scroll(-1, 1, iHeight)
		    line(iPos, 1)
		end if
	    elsif k = 329 then                  -- page up
		screen(max(iPos-iHeight, 1))
	    elsif k = 335 then                  -- end
		screen(iLines-(iHeight-1))
	    elsif k = 336 then                  -- down
		if iPos < iLines-(iHeight-1) then
		    iPos = iPos + 1
		    scroll(+1, 1, iHeight)
		    line(iPos+(iHeight-1), iHeight)
		end if
	    elsif k = 337 then                  -- page down
		screen(min(iPos+iHeight, iLines-(iHeight-1)))
	    elsif k >= 49 and k <= 57 then      -- 1-9
		if sBookmarks[k-48] > 0 then
		    screen(sBookmarks[k-48])
		end if
	    elsif k >= 376 and k <= 384 then    -- ALT 1-9
		sBookmarks[k-375] = iPos
	    --else
		--? k
	    end if
	end if
    end while
end procedure

main()
clear_screen()
